@@ -12,7 +12,11 @@
  * @param char *fmt - format of message to write (a la printf)
  * @param ... - other args for format string
  */
-void debug_printf(int level, char *fmt, ...);
+void debug_printf(int level, char *fmt, ...)
+#ifdef __GNUC__
+	__attribute__((format(printf, 2, 3)))
+#endif
+	;
 
 #endif
 
