@@ -8,6 +8,7 @@
 #include "strutl.h"
 
 #include <dirent.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
@@ -32,22 +33,6 @@ int nodequestioncomp(const void *pa, const void *pb) {
 }
 
 
-static char *unescapestr(const char *in)
-{
-    static char buf[8192];
-    if (in == 0) return 0;
-    strunescape(in, buf, sizeof(buf), 0);
-    return buf;
-}
-
-static char *escapestr(const char *in)
-{
-    static char buf[8192];
-    if (in == 0) return 0;
-    strescape(in, buf, sizeof(buf), 0);
-    return buf;
-}
-
 static void parse_variables(struct question *q, char *string)
 {
     char *wc, *owc;
@@ -150,6 +135,8 @@ static int rfc822db_template_initialize(struct template_db *db, struct configura
         return DC_NOTOK;
 
     dbdata->root = NULL;
+    dbdata->iterator = NULL;
+    dbdata->dirty = false;
     db->data = dbdata;
 
     return DC_OK;
@@ -174,7 +161,7 @@ static int rfc822db_template_load(struct template_db *db)
     if (path == NULL ||
         (inf = fopen(path, "r")) == NULL)
     {
-        INFO(INFO_VERBOSE, "Cannot open template file %s\n",
+        INFO(INFO_VERBOSE, "Cannot open template file %s",
             path ? path : "<empty>");
         return DC_NOTOK;
     }
@@ -188,7 +175,7 @@ static int rfc822db_template_load(struct template_db *db)
         name = rfc822_header_lookup(header, "name");
         if (name == NULL)
         {
-            INFO(INFO_ERROR, "Read a stanza without a name\n");
+            INFO(INFO_ERROR, "Read a stanza without a name");
             DELETE(header);
             continue;
         }
@@ -221,7 +208,7 @@ void rfc822db_template_dump(const void *node, const VISIT which, const int depth
     case postorder: 
     case leaf:
         p = t->lget((struct template *) t, NULL, "tag");
-        INFO(INFO_VERBOSE, "dumping template %s\n", p);
+        INFO(INFO_VERBOSE, "dumping template %s", p);
 
         for (field = template_fields_list; *field != NULL; field++)
         {
@@ -243,8 +230,14 @@ void rfc822db_template_dump(const void *node, const VISIT which, const int depth
             {
                 p = t->lget((struct template *) t, lang, *field);
                 if (p != NULL && p != t->lget((struct template *) t, NULL, *field))
-                    fprintf(outf, "%c%s-%s.UTF-8: %s\n",
-                        toupper((*field)[0]), (*field)+1, lang, escapestr(p));
+                {
+                    if (strcmp(lang, "C") == 0)
+                        fprintf(outf, "%c%s-C: %s\n",
+                            toupper((*field)[0]), (*field)+1, escapestr(p));
+                    else
+                        fprintf(outf, "%c%s-%s.UTF-8: %s\n",
+                            toupper((*field)[0]), (*field)+1, lang, escapestr(p));
+                }
             }
             lang = t->next_lang((struct template *) t, lang);
         }
@@ -257,10 +250,11 @@ static int rfc822db_template_save(struct template_db *db)
     struct template_db_cache *dbdata = db->data;
     char tmp[1024];
     const char *path;
+    struct stat st;
 
     if (outf != NULL)
     {
-            INFO(INFO_ERROR, "Internal incostisency error, outf is not NULL");
+            INFO(INFO_ERROR, "Internal inconsistency error, outf is not NULL");
             return DC_NOTOK;
     }
 
@@ -268,12 +262,19 @@ static int rfc822db_template_save(struct template_db *db)
     path = db->config->get(db->config, tmp, 0);
     if (path == NULL)
     {
-        INFO(INFO_ERROR, "Cannot open template file <empty>\n");
+        INFO(INFO_ERROR, "Cannot open template file <empty>");
         return DC_NOTOK;
     }
-    else if ((outf = fopen(path, "w")) == NULL)
+
+    if (!dbdata->dirty && stat(path, &st) == 0)
     {
-        INFO(INFO_ERROR, "Cannot open template file %s: %s\n",
+        INFO(INFO_DEBUG, "Template database %s clean; not saving", path);
+        return DC_OK;
+    }
+
+    if ((outf = fopen(path, "w")) == NULL)
+    {
+        INFO(INFO_ERROR, "Cannot open template file %s: %s",
             path, strerror(errno));
         return DC_NOTOK;
     }
@@ -306,29 +307,33 @@ static struct template *rfc822db_template_get(struct template_db *db,
 static int rfc822db_template_set(struct template_db *db, struct template *template)
 {
         struct template_db_cache *dbdata = db->data;
-    INFO(INFO_VERBOSE, "rfc822db_template_set(db,t=%s)\n", template->tag);
+    INFO(INFO_VERBOSE, "rfc822db_template_set(db,t=%s)", template->tag);
 
     tdelete(template, &dbdata->root, nodetemplatecomp);
     tsearch(template, &dbdata->root, nodetemplatecomp);
+    dbdata->dirty = true;
    
     template_ref(template);
 
     return DC_OK;
 }
 
-/* FIXME b0rken */
 static int rfc822db_template_remove(struct template_db *db, const char *tag)
 {
     struct template_db_cache *dbdata = db->data;
     struct template *t, t2;
 
-    INFO(INFO_VERBOSE, "rfc822db_template_remove(db,tag=%s)\n",tag);
+    INFO(INFO_VERBOSE, "rfc822db_template_remove(db,tag=%s)",tag);
 
+    memset(&t2, 0, sizeof (struct template));
     t2.tag = (char*) tag;
-    t = tdelete(&t, &dbdata->root, nodetemplatecomp);
+    t = tfind(&t2, &dbdata->root, nodetemplatecomp);
 
     if (t)
     {
+            t = *(struct template **) t;
+            tdelete(t, &dbdata->root, nodetemplatecomp);
+            dbdata->dirty = true;
             template_deref(t);
             return DC_OK;
     }
@@ -336,10 +341,57 @@ static int rfc822db_template_remove(struct template_db *db, const char *tag)
     return DC_NOTOK;
 }
 
-static struct template *rfc822db_template_iterate(struct template_db *db, void **iter)
+/* TODO: This is an ugly hack because there's no better way to do this
+ * within the constraints of twalk() (since there's no user-data argument).
+ * If we ever switch to some other tree API, this should go away
+ * immediately. If we ever need iterate() to be thread-safe, this *needs* to
+ * go away.
+ */
+di_slist *template_iterator;
+void rfc822db_template_makeiterator(const void *nodep, const VISIT which,
+    const int depth)
 {
-    INFO(INFO_VERBOSE, "rfc822db_template_iterate stub\n");
-    return NULL;
+    if (which == postorder || which == leaf)
+        di_slist_append(template_iterator,
+                        template_dup(*(struct template **) nodep));
+}
+
+void rfc822db_template_destroyiterator(void *data)
+{
+    template_deref((struct template *) data);
+}
+
+static struct template *rfc822db_template_iterate(struct template_db *db,
+    void **iter)
+{
+    struct template_db_cache *dbdata = db->data;
+    di_slist_node *node;
+    struct template *t;
+
+    INFO(INFO_VERBOSE, "rfc822db_template_iterate(db,*iter=%p)", *iter);
+
+    node = *(di_slist_node **) iter;
+    if (node == NULL) {
+        if (dbdata->iterator)
+            di_slist_destroy(dbdata->iterator,
+                             rfc822db_template_destroyiterator);
+        dbdata->iterator = di_slist_alloc();
+        template_iterator = dbdata->iterator; /* non-thread-safe */
+        twalk(dbdata->root, rfc822db_template_makeiterator);
+        template_iterator = NULL;
+        *iter = node = dbdata->iterator->head;
+    } else
+        *iter = node = node->next;
+
+    if (node == NULL) {
+        di_slist_destroy(dbdata->iterator, rfc822db_template_destroyiterator);
+        dbdata->iterator = NULL;
+        return NULL;
+    }
+
+    t = (struct template *) node->data;
+    template_ref(t);
+    return t;
 }
 
 /* config database */
@@ -353,6 +405,8 @@ static int rfc822db_question_initialize(struct question_db *db, struct configura
         return DC_NOTOK;
 
     dbdata->root = NULL;
+    dbdata->iterator = NULL;
+    dbdata->dirty = false;
     db->data = dbdata;
 
     return DC_OK;
@@ -381,7 +435,7 @@ static int rfc822db_question_load(struct question_db *db)
         if (errno == ENOENT)
             return DC_NOTOK;
 
-        INFO(INFO_VERBOSE, "Cannot open config database %s: %s\n",
+        INFO(INFO_VERBOSE, "Cannot open config database %s: %s",
             path ? path : "<empty>", strerror(errno));
         return DC_NOTOK;
     }
@@ -395,7 +449,7 @@ static int rfc822db_question_load(struct question_db *db)
 
         if (name == NULL || *name == 0)
         {
-            INFO(INFO_ERROR, "Read a stanza without a name\n");
+            INFO(INFO_ERROR, "Read a stanza without a name");
             DELETE(header);
             continue;
         }
@@ -435,11 +489,11 @@ void rfc822db_question_dump(const void *node, const VISIT which, const int depth
   case postorder: 
   case leaf:
 
-        INFO(INFO_VERBOSE, "dumping question %s\n", (q)->tag); 
+        INFO(INFO_VERBOSE, "dumping question %s", (q)->tag); 
         fprintf(outf, "Name: %s\n", escapestr((q)->tag));
         fprintf(outf, "Template: %s\n", escapestr((q)->template->tag));
-        if (((q)->flags & DC_QFLAG_SEEN) || (q)->value)
-            fprintf(outf, "Value: %s\n", ((q)->value ? escapestr((q)->value) : ""));
+        if ((q)->value)
+            fprintf(outf, "Value: %s\n", escapestr((q)->value));
 
         if ((owner = (q)->owners))
         {
@@ -487,17 +541,25 @@ static int rfc822db_question_save(struct question_db *db)
     struct question_db_cache *dbdata = db->data;
     const char *path;
     char tmp[1024];
+    struct stat st;
     
     snprintf(tmp, sizeof(tmp), "%s::path", db->configpath);
     path = db->config->get(db->config, tmp, 0);
     if (path == NULL)
     {
-        INFO(INFO_ERROR, "Cannot open question file <empty>\n");
+        INFO(INFO_ERROR, "Cannot open question file <empty>");
         return DC_NOTOK;
     }
-    else if ((outf = fopen(path, "w")) == NULL)
+
+    if (!dbdata->dirty && stat(path, &st) == 0)
     {
-        INFO(INFO_ERROR, "Cannot open question file %s: %s\n",
+        INFO(INFO_DEBUG, "Question database %s clean; not saving", path);
+        return DC_OK;
+    }
+
+    if ((outf = fopen(path, "w")) == NULL)
+    {
+        INFO(INFO_ERROR, "Cannot open question file %s: %s",
             path, strerror(errno));
         return DC_NOTOK;
     }
@@ -535,10 +597,22 @@ static int rfc822db_question_set(struct question_db *db, struct question *questi
     struct question_db_cache *dbdata = db->data;
     struct question *q;
 
-    INFO(INFO_VERBOSE, "rfc822db_question_set(db,q=%s,q=%p)\n", question->tag, question);
+    INFO(INFO_VERBOSE, "rfc822db_question_set(db,q=%s,q=%p)", question->tag, question);
+
+    /* merge old and new owner lists */
+    q = tfind(question, &dbdata->root, nodequestioncomp);
+    if (q != NULL) {
+        q = *(struct question **)q;
+        const struct questionowner *owner = q->owners;
+        while (owner) {
+            question_owner_add(question, owner->owner);
+            owner = owner->next;
+        }
+    }
 
     tdelete(question, &dbdata->root, nodequestioncomp);
     q = tsearch(question, &dbdata->root, nodequestioncomp);
+    dbdata->dirty = true;
     question_ref(question);
 
     return DC_OK;
@@ -559,11 +633,77 @@ static int rfc822db_question_disown(struct question_db *db, const char *tag,
     return DC_OK;
 }
 
+static int rfc822db_question_remove(struct question_db *db, const char *tag)
+{
+    struct question_db_cache *dbdata = db->data;
+    struct question *q, q2;
+
+    INFO(INFO_VERBOSE, "rfc822db_question_remove(db,tag=%s)", tag);
+
+    memset(&q2, 0, sizeof (struct question));
+    q2.tag = (char*) tag;
+    q = tfind(&q2, &dbdata->root, nodequestioncomp);
+    if (q != NULL) {
+        q = *(struct question **) q;
+        tdelete(q, &dbdata->root, nodequestioncomp);
+        dbdata->dirty = true;
+        question_deref(q);
+        return DC_OK;
+    }
+    return DC_NOTOK;
+}
+
+/* TODO: This is an ugly hack because there's no better way to do this
+ * within the constraints of twalk() (since there's no user-data argument).
+ * If we ever switch to some other tree API, this should go away
+ * immediately. If we ever need iterate() to be thread-safe, this *needs* to
+ * go away.
+ */
+di_slist *question_iterator;
+void rfc822db_question_makeiterator(const void *nodep, const VISIT which,
+    const int depth)
+{
+    if (which == postorder || which == leaf)
+        di_slist_append(question_iterator,
+                        question_dup(*(struct question **) nodep));
+}
+
+void rfc822db_question_destroyiterator(void *data)
+{
+    question_deref((struct question *) data);
+}
+
 static struct question *rfc822db_question_iterate(struct question_db *db,
     void **iter)
 {
-        INFO(INFO_VERBOSE, "rfc822db_question_iterate stub\n");
+    struct question_db_cache *dbdata = db->data;
+    di_slist_node *node;
+    struct question *q;
+
+    INFO(INFO_VERBOSE, "rfc822db_question_iterate(db,*iter=%p)", *iter);
+
+    node = *(di_slist_node **) iter;
+    if (node == NULL) {
+        if (dbdata->iterator)
+            di_slist_destroy(dbdata->iterator,
+                             rfc822db_question_destroyiterator);
+        dbdata->iterator = di_slist_alloc();
+        question_iterator = dbdata->iterator; /* non-thread-safe */
+        twalk(dbdata->root, rfc822db_question_makeiterator);
+        question_iterator = NULL;
+        *iter = node = dbdata->iterator->head;
+    } else
+        *iter = node = node->next;
+
+    if (node == NULL) {
+        di_slist_destroy(dbdata->iterator, rfc822db_question_destroyiterator);
+        dbdata->iterator = NULL;
         return NULL;
+    }
+
+    q = (struct question *) node->data;
+    question_ref(q);
+    return q;
 }
 
 struct template_db_module debconf_template_db_module = {
@@ -583,6 +723,7 @@ struct question_db_module debconf_question_db_module = {
     set: rfc822db_question_set,
     get: rfc822db_question_get,
     disown: rfc822db_question_disown,
+    remove: rfc822db_question_remove,
     iterate: rfc822db_question_iterate,
 };
 
