@@ -10,7 +10,7 @@
  * friendly implementation. I've taken care to make the prompts work well
  * with screen readers and the like.
  *
- * $Id: text.c 26101 2005-03-06 22:07:18Z joeyh $
+ * $Id: text.c 34558 2006-01-27 01:35:19Z cjwatson $
  *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
@@ -44,6 +44,7 @@
 #include "question.h"
 #include "frontend.h"
 #include "database.h"
+#include "plugin.h"
 #include "strutl.h"
 
 #include <ctype.h>
@@ -59,11 +60,11 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
-#define q_get_extended_description(q)   question_get_field((q), "", "extended_description")
-#define q_get_description(q)		question_get_field((q), "", "description")
-#define q_get_choices(q)		question_get_field((q), "", "choices")
-#define q_get_choices_vals(q)		question_get_field((q), NULL, "choices")
-#define q_get_indices(q)		question_get_field((q), "", "indices")
+struct frontend_data {
+	char *previous_title;
+};
+
+typedef int (text_handler)(struct frontend *obj, struct question *q);
 
 #define MAKE_UPPER(C) do { if (islower((int) C)) { C = (char) toupper((int) C); } } while(0)
 
@@ -108,7 +109,7 @@ strpad (char *what, size_t maxsize)
  * Description: get the width of the current terminal
  * Assumptions: doesn't handle resizing; caches value on first call
  */
-static const int getwidth(void)
+static int getwidth(void)
 {
 	static int res = 80;
 	static int inited = 0;
@@ -128,22 +129,6 @@ static const int getwidth(void)
 	return res;
 }
 
-/*
- * Function: get_text
- * Input: struct frontend *obj - frontend object
- *        const char *template - template name
- *        const char *fallback - string to use if not available
- * Output: const char * - ptr to string, translated if possible
- * Description: get the translated version of a string
- * Assumptions: None.
- */
-static const char *
-get_text(struct frontend *obj, const char *template, const char *fallback)
-{
-	struct question *q = obj->qdb->methods.get(obj->qdb, template);
-	return q ? q_get_description(q) : fallback;
-}
-
 /*
  * Function: wrap_print
  * Input: const char *str - string to display
@@ -182,7 +167,7 @@ static void text_handler_displaydesc(struct frontend *obj, struct question *q)
 	    strcmp(q->template->type, "error") == 0)
 	{
 		if (strcmp(q->template->type, "error") == 0)
-			printf(get_text(obj, "debconf/text-error", "!! ERROR: %s"), descr);
+			printf(question_get_text(obj, "debconf/text-error", "!! ERROR: %s"), descr);
 		else
 			printf("%s", descr);
 		printf("\n\n");
@@ -210,18 +195,18 @@ static void
 show_help (struct frontend *obj, struct question *q)
 {
 	char *descr = q_get_description(q);
-	printf("%s", get_text(obj, "debconf/text-help-keystrokes", "KEYSTROKES:"));
+	printf("%s", question_get_text(obj, "debconf/text-help-keystrokes", "KEYSTROKES:"));
 	printf("\n  %c  ", CHAR_HELP);
-	printf("%s", get_text(obj, "debconf/text-help-help", "Display this help message"));
+	printf("%s", question_get_text(obj, "debconf/text-help-help", "Display this help message"));
 	if (obj->methods.can_go_back (obj, q)) {
 		printf("\n  %c  ", CHAR_GOBACK);
-		printf("%s", get_text(obj, "debconf/text-help-goback", "Go back to previous question"));
+		printf("%s", question_get_text(obj, "debconf/text-help-goback", "Go back to previous question"));
 	}
 	if (strcmp(q->template->type, "string") == 0 ||
 	    strcmp(q->template->type, "passwd") == 0 ||
 	    strcmp(q->template->type, "multiselect") == 0) {
 		printf("\n  %c  ", CHAR_CLEAR);
-		printf("%s", get_text(obj, "debconf/text-help-clear", "Select an empty entry"));
+		printf("%s", question_get_text(obj, "debconf/text-help-clear", "Select an empty entry"));
 	}
 	printf("\n");
 	wrap_print(descr);
@@ -356,14 +341,14 @@ static int text_handler_boolean(struct frontend *obj, struct question *q)
 	}
 
 	do {
-		printf("  %d. %s%s", 1, get_text(obj, "debconf/yes", "Yes"), (1 == def ? " [*]" : "    "));
-		printf("  %d. %s%s", 2, get_text(obj, "debconf/no", "No"), (2 == def ? " [*]" : ""));
+		printf("  %d. %s%s", 1, question_get_text(obj, "debconf/yes", "Yes"), (1 == def ? " [*]" : "    "));
+		printf("  %d. %s%s", 2, question_get_text(obj, "debconf/no", "No"), (2 == def ? " [*]" : ""));
 		printf("\n");
 		if (def)
-			printf(get_text(obj, "debconf/text-prompt-default",
+			printf(question_get_text(obj, "debconf/text-prompt-default",
 					"Prompt: '%c' for help, default=%d> "), CHAR_HELP, def);
 		else
-			printf(get_text(obj, "debconf/text-prompt",
+			printf(question_get_text(obj, "debconf/text-prompt",
 					"Prompt: '%c' for help> "), CHAR_HELP);
 		get_answer(buf, sizeof(buf));
 		if (buf[0] == CHAR_HELP && buf[1] == 0)
@@ -410,6 +395,10 @@ static int text_handler_multiselect(struct frontend *obj, struct question *q)
 		return DC_NOTOK;
 	choices = malloc(sizeof(char *) * count);
 	choices_translated = malloc(sizeof(char *) * count);
+	for (i = 0; i < count; ++i)
+	{
+		choices[i] = choices_translated[i] = NULL;
+	}
 	tindex = malloc(sizeof(int) * count);
 	if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_translated, tindex, count) != count)
 	{
@@ -443,7 +432,7 @@ static int text_handler_multiselect(struct frontend *obj, struct question *q)
 
   DISPLAY:
 	printlist (obj, q, count, choices_translated, tindex, selected);
-	printf(get_text(obj, "debconf/text-prompt-default-string", 
+	printf(question_get_text(obj, "debconf/text-prompt-default-string", 
 		"Prompt: '%c' for help, default=%s> "), CHAR_HELP, defval);
 	get_answer(answer, sizeof(answer));
 	if (answer[0] == CHAR_HELP && answer[1] == 0)
@@ -533,6 +522,10 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 		return DC_NOTOK;
 	choices = malloc(sizeof(char *) * count);
 	choices_translated = malloc(sizeof(char *) * count);
+	for (i = 0; i < count; ++i)
+	{
+		choices[i] = choices_translated[i] = NULL;
+	}
 	tindex = malloc(sizeof(int) * count);
 	if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_translated, tindex, count) != count)
 	{
@@ -559,11 +552,11 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 	do {
 		printlist (obj, q, count, choices_translated, tindex, selected);
 		if (def >= 0 && choices_translated[def]) {
-			printf(get_text(obj, "debconf/text-prompt-default", 
+			printf(question_get_text(obj, "debconf/text-prompt-default", 
 				"Prompt: '%c' for help, default=%d> "),
 					CHAR_HELP, def+1);
 		} else {
-			printf(get_text(obj, "debconf/text-prompt",
+			printf(question_get_text(obj, "debconf/text-prompt",
 				"Prompt: '%c' for help> "), CHAR_HELP);
 		}
 		get_answer(answer, sizeof(answer));
@@ -612,8 +605,9 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 static int text_handler_note(struct frontend *obj, struct question *q)
 {
 	char buf[100] = {0};
-	printf("%s\n", get_text(obj, "debconf/cont-prompt",
-				"[Press enter to continue]"));
+	printf("%s ", question_get_text(obj, "debconf/cont-prompt",
+			       "[Press enter to continue]"));
+	fflush(stdout);
 	while (1)
 	{
 		get_answer(buf, sizeof(buf));
@@ -695,9 +689,9 @@ static int text_handler_string(struct frontend *obj, struct question *q)
 	const char *defval = question_getvalue(q, "");
 	while (1) {
 		if (defval)
-			printf(get_text(obj, "debconf/text-prompt-default-string", "Prompt: '%c' for help, default=%s> "), CHAR_HELP, defval);
+			printf(question_get_text(obj, "debconf/text-prompt-default-string", "Prompt: '%c' for help, default=%s> "), CHAR_HELP, defval);
 		else
-			printf(get_text(obj, "debconf/text-prompt", "Prompt: '%c' for help> "), CHAR_HELP);
+			printf(question_get_text(obj, "debconf/text-prompt", "Prompt: '%c' for help> "), CHAR_HELP);
 		fflush(stdout);
 		get_answer(buf, sizeof(buf));
 		if (buf[0] == CHAR_HELP && buf[1] == 0)
@@ -748,7 +742,7 @@ static int text_handler_error(struct frontend *obj, struct question *q)
 /* ----------------------------------------------------------------------- */
 struct question_handlers {
 	const char *type;
-	int (*handler)(struct frontend *obj, struct question *q);
+	text_handler *handler;
 } question_handlers[] = {
 	{ "boolean",	text_handler_boolean },
 	{ "multiselect", text_handler_multiselect },
@@ -758,10 +752,11 @@ struct question_handlers {
 	{ "string",	text_handler_string },
 	{ "text",	text_handler_text },
 	{ "error",	text_handler_error },
+	{ "",		NULL },
 };
 
 /*
- * Function: text_intitialize
+ * Function: text_initialize
  * Input: struct frontend *obj - frontend UI object
  *        struct configuration *cfg - configuration parameters
  * Output: int - DC_OK, DC_NOTOK
@@ -773,6 +768,9 @@ struct question_handlers {
  */
 static int text_initialize(struct frontend *obj, struct configuration *conf)
 {
+	struct frontend_data *data = NEW(struct frontend_data);
+	data->previous_title = NULL;
+	obj->data = data;
 	obj->interactive = 1;
 	signal(SIGINT, SIG_IGN);
 	return DC_OK;
@@ -786,7 +784,7 @@ static int text_initialize(struct frontend *obj, struct configuration *conf)
  * Description: tells whether confmodule supports backing up
  * Assumptions: none
  */
-static int
+static bool
 text_can_go_back(struct frontend *obj, struct question *q)
 {
 	return (obj->capability & DCF_CAPB_BACKUP);
@@ -801,32 +799,73 @@ text_can_go_back(struct frontend *obj, struct question *q)
  */
 static int text_go(struct frontend *obj)
 {
+	struct frontend_data *data = (struct frontend_data *) obj->data;
 	struct question *q = obj->questions;
 	int i;
 	int ret = DC_OK;
-	int display_title = 1;
 
 	while (q != NULL) {
 		for (i = 0; i < DIM(question_handlers); i++) {
-			if (strcmp(q->template->type, question_handlers[i].type) == 0) 
-			{
+			text_handler *handler;
+			struct plugin *plugin = NULL;
 
-				if (display_title)
+			if (*question_handlers[i].type)
+				handler = question_handlers[i].handler;
+			else {
+				plugin = plugin_find(obj, q->template->type);
+				if (plugin) {
+					INFO(INFO_DEBUG, "Found plugin for %s", q->template->type);
+					handler = (text_handler *) plugin->handler;
+				} else {
+					INFO(INFO_DEBUG, "No plugin for %s", q->template->type);
+					continue;
+				}
+			}
+
+			if (plugin || strcmp(q->template->type, question_handlers[i].type) == 0)
+			{
+				if (!data->previous_title ||
+				    strcmp(obj->title, data->previous_title) != 0)
 				{
-					printf("%s\n\n", obj->title);
-					display_title = 0;
+					size_t underline_len;
+					char *underline;
+
+					/* TODO: can't tell if we called go()
+					 * twice during one progress bar, but
+					 * I'm guessing that's relatively
+					 * unimportant (c.f. #271707).
+					 */
+					if (obj->progress_title != NULL)
+						putchar('\n');
+					underline_len = strlen(obj->title);
+					underline = malloc(underline_len + 1);
+					memset(underline, '-', underline_len);
+					underline[underline_len] = '\0';
+					printf("%s\n%s\n\n", obj->title, underline);
+					free(underline);
+					if (data->previous_title)
+						free(data->previous_title);
+					data->previous_title = strdup(obj->title);
 				}
 				text_handler_displaydesc(obj, q);
-				ret = question_handlers[i].handler(obj, q);
+				ret = handler(obj, q);
+				putchar('\n');
 				if (ret == DC_OK)
 					obj->qdb->methods.set(obj->qdb, q);
 				else if (ret == DC_GOBACK && q->prev != NULL)
 					q = q->prev;
-				else
+				else {
+					if (plugin)
+						plugin_delete(plugin);
 					return ret;
+				}
+				if (plugin)
+					plugin_delete(plugin);
 				break;
 			}
 		}
+		if (i == DIM(question_handlers))
+			return DC_NOTIMPL;
 		if (ret == DC_OK)
 			q = q->next;
 	}
@@ -845,7 +884,7 @@ static void text_progress_start(struct frontend *obj, int min, int max, const ch
 	fflush(stdout);
 }
 
-static void text_progress_set(struct frontend *obj, int val)
+static int text_progress_set(struct frontend *obj, int val)
 {
 	static int last = 0;
 	int new;
@@ -857,17 +896,20 @@ static void text_progress_set(struct frontend *obj, int val)
 		last = 0;
 	/*  Prevent verbose output  */
 	if (new / 10 == last / 10)
-		return;
+		return DC_OK;
 	last = new;
 	printf("..%d%%", new);
 	fflush(stdout);
+
+	return DC_OK;
 }
 
 static void text_progress_stop(struct frontend *obj)
 {
-	INFO(INFO_DEBUG, "%s\n", __FUNCTION__);
+	INFO(INFO_DEBUG, "%s", __FUNCTION__);
 	printf("\n");
 	fflush(stdout);
+	DELETE(obj->progress_title);
 }
 
 struct frontend_module debconf_frontend_module =
